module Interpreter.Parser where

import Common.Utils

import DataStructures.Tree

import InternalRepresentation.Tokens
import InternalRepresentation.Values

import SyntaxEngine.Parsing
import SyntaxEngine.Result


type FunParser = Parser Token PType


pDefinition :: FunParser
pDefinition [] = Failure "Nothing to parse"
pDefinition toks = p_seq Definition [pSymbol, pKeyword "=", pExpr] toks

pSymbol :: FunParser
pSymbol (t:ts) = case t of
	TSymbol _ -> Success (leaf t, ts)
	otherwise -> Failure "Symbol expected."

pKeyword :: String -> FunParser
pKeyword key (t:ts) = case t of
	TKeyword k | k == key -> Success (leaf t, ts)
	otherwise    -> Failure (key ++ " expected instead of " ++ show t)

pDelimiter :: String -> FunParser
pDelimiter d (t:ts) = case t of
	TDelimiter d -> Success (leaf t, ts)
	otherwise    -> Failure (d ++ " expected.")


pValue :: FunParser
pValue (t@(TValue _):ts) = Success (leaf t, ts)
pValue otherwise = Failure "Value expected."

--pSubExpr :: FunParser
--pSubExpr :: p_seq SubExpr [pDelimiter "(", pExpr, pDelimiter ")"]

pExpr :: FunParser
pExpr = pExpr' [["+","-"],["*","/"],["^"]] where 
	
	pSimpleExpr :: FunParser
	pSimpleExpr = p_alt [pValue, pSymbol] "Failed to parse expression." -- <or> pSubExpr

	pExpr' :: [[String]] -> FunParser
	pExpr' []  tokens = pSimpleExpr tokens
        pExpr' ops tokens = pSimpleExpr tokens `feed_into` pAfterExpr where
	
		pAfterExpr lhs []     = Success (lhs, [])
		pAfterExpr lhs (t:ts) = case t of
			TSymbol s | s `elem` (concat ops) -> lhs_op_rhs `feed_into` pAfterExpr 
			otherwise -> Success (lhs, t:ts)
			where
				op = leaf t
				rhsRes = pExpr' (higherPriority (getSymbol t) ops) ts
				lhs_op_rhs = case rhsRes of
					Failure _ -> rhsRes
					Success (rhs, rem) -> Success (node FunCall [op, lhs, rhs], rem)
				
				higherPriority :: String -> [[String]] -> [[String]]
				higherPriority op ops = case ops of
					[] -> []
					o:os | op `elem` o -> os
					o:os | otherwise   -> higherPriority op os
				
				getSymbol (TSymbol s) = s








data PType
	= FunCall
	| FunDef
	| SubExpr
	| Definition
	deriving (Eq, Show)

type PNode = ParseNode Token PType
leaf :: Token -> Tree PNode
leaf token = Tree (Leaf token) []

node :: PType -> Forest PNode -> Tree PNode
node ptype children = Tree (Node ptype) children

p_seq :: PType -> [FunParser] -> FunParser
p_seq ptype parsers tokens = p_seq' parsers ([], tokens) where
	p_seq' [] (parsed, rem) = Success (node ptype (reverse parsed), rem)
	p_seq' (p:ps) (parsed, rem) = case p rem of
		Success (pn, rem') -> p_seq' ps (pn : parsed, rem')
		Failure f -> Failure f


p_alt :: [FunParser] -> String -> FunParser
p_alt [] errStr tokens = Failure errStr
p_alt (p:ps) errStr tokens = case p tokens of
	Success s -> Success s
	Failure f -> p_alt ps errStr tokens



feed_into :: (Result (Tree PNode, [Token]) String) -> (Tree PNode -> FunParser) -> (Result (Tree PNode, [Token]) String)
feed_into result parser = case result of
	Success (pnode, rem) -> parser pnode rem
	Failure f            -> Failure f





{-

Rules:

map parse_line Normal >>> recursive_check Normal

1. If line successful parsed, parse nested lines
2. What can be parsed inside depends on header line
3. Check line sequence order rules / limitations

the last thing should be WithPos ? in order to keep track of the position
Tree
	leaf == withPos token
	node == parseNode

To get a more readable thing, extractors can be used to get cleaned version.

Two steps:
grabbers : tokens => consumed + what is it
extractors : what is it => cleaned version

What is definition grabber?

	symbol <then> keyword = <then> expr


expr
	fundef <or> funcall <or> value <or> tuple expr <or> symbol <or> list expr <or> set expr



pExpr :: Parser Expr

pValue :: Parser Value

pDef :: Parser Definition

pFunCall :: 

pFunDef

pSymbol

pIf






data Definition 
	= Definition {
		symbol :: WithPos Symbol
		expr :: WithPos Expr
		}

data Expr
	= FunCall
		symbol :: WithPos Symbol







-}
