{-
Tokenizing process
------------------

1. Transorm the text into a tree of lines

2. Tokenize each line independently

3. If they are errors, summerize all errors. If all is correct, return the token tree.

Step 2: Tokenize a line
-----------------------
A few things to consider:	
	a. dots '.' cannot have neightboring white spaces
	b. value tokens cannot appear consecutive
	c. after a decimal number, a dot is forbidden
	d. after a dot, don't parse decimals but ints

-}

module Interpreter.Tokenizer where

import Common.Utils

import DataStructures.Tree

import SyntaxEngine.Tokenizing
import SyntaxEngine.Result
import SyntaxEngine.Pos

import InternalRepresentation.Tokens
import InternalRepresentation.Values





type Line = (Int, String)

delimiters = "()[]{},;" -- (the dot has a special status)


-------------------------------------------------------------------------------------
-- Some shortcuts to make tokens from string
-- (Beware: string must be "correct", no safety checks!)
-------------------------------------------------------------------------------------

makeChar str = case str of
	'\'':c:'\'':[] -> (Token . TValue . VChar) c
	"'\\n'" -> (Token . TValue . VChar) '\n'
	"'\\t'" -> (Token . TValue . VChar) '\t'
	otherwise -> Error "Char expected, incorrectly formated."

makeString str = if (head str == '"') && (last str == '"')
	then (Token . TValue . VString) str
	else Error "Missing ending quotes."

makeInt str = (Token . TValue) (VInt (read str))
makeDecimal str = (Token . TValue) (VFloat (read str))

makeDate str = (Token . TValue) (VDate (subToInt 1 4 str) (subToInt 6 7 str) (subToInt 9 10 str))
makeTime str = (Token . TValue) (VTime (subToInt 1 2 str) (subToInt 4 5 str) (subToInt 7 8 str))

subToInt :: Int -> Int -> (String -> Int)
subToInt from to = read . take (to - from + 1) . drop (from - 1)

makeDelimiter str = (Token . TDelimiter) str

makeKeyword str = (Token . TKeyword) str

makeSymbol str = (Token . TSymbol) str

makeComment str = Ignored

makeWhiteSpace str = Ignored

makeError str = Error str

-------------------------------------------------------------------------------------
-- Tries to grab a token from the front of the string
-------------------------------------------------------------------------------------

type TokenGrabber = String -> Maybe (String, TokRes Token)

tryGrabbing :: String -> (String -> a) -> Maybe (String, a)
tryGrabbing str fun = case str of
	"" -> Nothing
	_  -> Just (str, fun str)

grabWhiteSpace str = tryGrabbing (grab "[ \t]+" str) makeWhiteSpace

grabDelimiter delim (c:cs) | c `elem` delim = Just (c:[], makeDelimiter (c:[]))
grabDelimiter delim (c:cs) | otherwise           = Nothing

grabComment str = tryGrabbing (grab "#.*" str) makeComment

grabDecimal str = tryGrabbing (grab "-?[0-9]+\\.[0-9]+" str) makeDecimal

grabAtom ('\'':cs) = tryGrabbing ('\'' : take_until "'" cs) makeChar
grabAtom ('"' :cs) = tryGrabbing ('"' : take_until "\"" cs) makeString
grabAtom str       = tryGrabbing (take_while_not_in (delimiters ++ " \t#") str) makeAtom where
	makeAtom str 
		| str == "="    = makeKeyword str
		| str == "if"   = makeKeyword str
		| str == "else" = makeKeyword str
		| str `matches` "[0-9]{4}-[0-9]{2}-[0-9]{2}" = makeDate str
		| str `matches` "[0-9]{4}/[0-9]{2}/[0-9]{2}" = makeDate str
		| str `matches` "-?[0-9]+" = makeInt str
		| str `matches` "-?[0-9]+.*" = makeError ("Invalid token: " ++ str ++  ". Number or date expected.")
		| otherwise = makeSymbol str

-------------------------------------------------------------------------------------
-- What can be tokenized after what
-------------------------------------------------------------------------------------

data TokenizerMode = Normal | AfterAtom | AfterDot | Stop

nextMode :: TokRes Token -> TokenizerMode
nextMode tokRes = case tokRes of
	Token (TValue _)       -> AfterAtom
	Token (TDelimiter ".") -> AfterDot
	Error _                -> Stop
	otherwise              -> Normal


grabbers :: TokenizerMode -> [TokenGrabber]
grabbers Normal    = [grabWhiteSpace, grabDelimiter "()[]{},;" , grabComment, grabDecimal, grabAtom]
grabbers AfterAtom = [grabWhiteSpace, grabDelimiter ")]}."                                         ]
grabbers AfterDot  = [                grabDelimiter "([{",                                 grabAtom]
grabbers Stop      = []


-------------------------------------------------------------------------------------
-- Single line meta-tokenizer
-- type MetaTokenizer token = String -> [(String, TokRes token)]
-------------------------------------------------------------------------------------



mtokenize :: TokenizerMode -> MetaTokenizer Token
mtokenize _    ""  = []
mtokenize mode str = (consumed, tokenRes) : mtokenize next_mode remaining where
	remaining = drop (length consumed) str
	(consumed, tokenRes) = try_alt (grabbers mode) (str, Error ("Unexpected character(s): " ++ str)) str
	next_mode = nextMode tokenRes
	

	
-------------------------------------------------------------------------------------
-- Text -> forest of lines
-------------------------------------------------------------------------------------

toLineForest :: String -> Forest (Int, String)
toLineForest = lines >>> indexed 1 >>> filter (not . isBlank) >>> toForest isIndented where
	isBlank (_, str) = str `matches` "[ \t]*$"
	isIndented a b = countTabs b > countTabs a
	countTabs (i, str) = length (takeWhile (== '\t') str)




tokenize :: String -> Result (Forest [WithPos Token]) [ErrorMsg]
tokenize text = ans where
	forest = toLineForest text
	
	tokenizeStream :: Tokenizer Token
	tokenizeStream = makeTokenizer (mtokenize Normal)
	
	tokenizeLine (i, str) = tokenizeStream (Pos i 1, str)

	tokenForest = fMap tokenizeLine forest

	getTokens (Success tokens) = tokens
	getErrors (Failure err)    = err

	ans = case failures (fFlatten tokenForest) of
		[]   -> Success $ fMap getTokens tokenForest
		errs -> Failure $ concat errs

