module SyntaxEngine.Tokenizing where

import Common.Utils

import SyntaxEngine.Pos
import SyntaxEngine.Result

import Text.Regex.Posix


-------------------------------------------------------------------------------------
-- Utility functions
-------------------------------------------------------------------------------------

{-
Checks wether the string matches the regular expression pattern.

In:
	"aabbcc" "a+b*(cc)?"
Out:
	True

In:
	"aabbcc" "a*"
Out:
	False
-}
matches :: String -> String -> Bool
matches str pattern = str =~ ("^" ++ pattern)

{-
Retruns the longest sub-string matching the regular expression.

In: 
	"a*b+" "aabbcc"
Out:
	"aabb"

In:
	"c+" "aabbcc"
Out:
	""
-}
grab :: String -> String -> String
grab pattern str = str =~ ("^(" ++ pattern ++ ")")



take_while_in     set (c:cs) = c : (takeWhile (\ c -> c `elem` set) cs)
take_while_not_in set (c:cs) = c : (takeWhile (\ c -> not (c `elem` set)) cs)
take_until        set (c:cs) = c : grab_until' cs where
	grab_until' "" = ""
	grab_until' (c:cs) | c `elem` set = [c]
	grab_until' (c:cs) | otherwise = c : grab_until' cs


-------------------------------------------------------------------------------------
-- Interface types
-------------------------------------------------------------------------------------


type Stream = (Pos, String)


type Tokenizer token = Stream -> Result [WithPos token] [ErrorMsg]


type MetaTokenizer token = String -> [(String, TokRes token)]


data TokRes token
	= Token token
	| Ignored
	| Error String


isError (Error _) = True
isError  _        = False

isIgnored Ignored = True
isIgnored  _      = False

isToken (Token _) = True
isToken  _        = False


-------------------------------------------------------------------------------------
-- MetaTokenizer -> Tokenizer
--
-- this adds positional data and summarizes the result
-------------------------------------------------------------------------------------


makeTokenizer :: MetaTokenizer token -> Tokenizer token
makeTokenizer metaTok (pos, string) = ans where
	str_tokRes = metaTok string
	wp_tokRes = withPos pos str_tokRes where
		withPos pos [] = []
		withPos pos ((str, tokRes):ts) = tokResWP : withPos pos_after ts where
			tokResWP = WithPos {
				inside = tokRes,
				start = pos,
				end = shift pos (length str - 1)
				}
			pos_after = shift pos (length str)
	
	ans = case filter (isError . inside) wp_tokRes of
		[]   -> Success (mapInside getToken (removeIgnored wp_tokRes))
		errs -> Failure (mapInside getError errs)
		where
			removeIgnored = filter (not . isIgnored . inside)
			getToken (Token tok) = tok
			getError (Error err) = err
			
		